local _, addon = ...
local lib = addon.libraries.TomCatsLibs.Events
local eventFrame = CreateFrame("Frame")
local eventListeners = { }

local function validateRegisterUnregisterInput(event, listener)
    if ((not event) or (type(event) ~= "string")) then error("Event must be specified", 3) end
    if ((not listener) or (type(listener) ~= "table" and type(listener) ~= "function")) then error("Listener must be specified", 3) end
end

-- todo: Make it ok to omit the listener and assume addon is the listener
function lib.RegisterEvent(event, listener)
    validateRegisterUnregisterInput(event, listener)
    eventListeners[event] = eventListeners[event] or {}
    for i = 1, #eventListeners[event] do
        if (eventListeners[event][i] == listener) then error(tostring(listener) .. " already associated with " .. event, 2) end
    end
    table.insert(eventListeners[event], listener)
    if (#eventListeners[event] == 1) then
        eventFrame:RegisterEvent(event)
    end
end

function lib.UnregisterEvent(event, listener)
    validateRegisterUnregisterInput(event, listener)
    for i = 1, #eventListeners[event] do
        if (eventListeners[event][i] == listener) then
            table.remove(eventListeners[event],i)
            if (#eventListeners[event] == 0) then
                eventFrame:UnregisterEvent(event);
            end
            break;
        end
    end
end

-- todo: make these late binding validations include a previously captured debugstack from when the event was registered
local function eventHandler(_, event, ...)
    local eventListenersQueue = Mixin({}, eventListeners[event])
    if (eventListenersQueue) then
        for i = 1, #eventListenersQueue do
            if (type(eventListenersQueue[i]) == "table") then
                if (eventListenersQueue[i][event] and (type(eventListenersQueue[i][event]) == "function")) then
                    eventListenersQueue[i][event](eventListenersQueue[i], event, ...)
                    return
                else
                    error("cannot find a function named " .. event .. " in table " .. tostring(eventListenersQueue[i]))
                end
            end
            if (type(eventListenersQueue[i]) == "function") then
                -- todo: check that the listener is still registered in the original listeners table
                eventListenersQueue[i](event, ...)
                return
            else
                error("cannot relay the event for " .. event)
            end
        end
    end
end

eventFrame:SetScript("OnEvent", eventHandler)
